#include <ode/ode.h>
#include <drawstuff/drawstuff.h>

static dWorldID world;
static dSpaceID space;
static dJointGroupID contactgroup;

static dBodyID body1;
static dGeomID geom1; 




#ifdef dDOUBLE
#define dsDrawBox dsDrawBoxD
#define dsDrawSphere dsDrawSphereD
#define dsDrawCylinder dsDrawCylinderD
#define dsDrawCapsule dsDrawCapsuleD
#define dsDrawConvex dsDrawConvexD
#define dsDrawTriangle dsDrawTriangleD
#endif



#define HFIELD_WIDTH			REAL( 10.0 )
#define HFIELD_DEPTH			REAL( 10.0 )

// Our heightfield geom
dGeomID gheight;

unsigned char* height_map;
int height_xsize;
int height_ysize;

// rbg}bv獂}bv𐶐
const unsigned char* createHeightMapForBitmap( const char* file_name, int* x_size, int* y_size )
{
    unsigned char* height_map = NULL;

    unsigned short  bfType;
    unsigned long bfSize;
    unsigned short  bfReserved1;
    unsigned short  bfReserved2;
    unsigned long bfOffBits  ;
    unsigned long biSize;
    long  biWidth;
    long  biHeight;
    unsigned short  biPlanes;
    unsigned short  biBitCount;
    unsigned long biCompression;
    unsigned long biSizeImage;
    long  biXPelsPerMeter;
    long  biYPelsPerMeter;
    unsigned long biClrUsed;
    unsigned long biClrImportant;

    FILE *fp;
    fp = fopen( file_name, "rb" );
    if (fp == NULL) {
        printf("can not file open : %s\n", file_name );
        return NULL;
    }

    fread( &bfType, sizeof(unsigned short), 1, fp );
    fread( &bfSize, sizeof(unsigned long), 1, fp );
    fread( &bfReserved1, sizeof(unsigned short), 1, fp );
    fread( &bfReserved2, sizeof(unsigned short), 1, fp );
    fread( &bfOffBits, sizeof(unsigned long), 1, fp );
    fread( &biSize, sizeof(unsigned long), 1, fp );
    fread( &biWidth, sizeof(long), 1, fp );
    fread( &biHeight, sizeof(long), 1, fp );
    fread( &biPlanes, sizeof(unsigned short), 1, fp );
    fread( &biBitCount, sizeof(unsigned short), 1, fp );
    fread( &biCompression, sizeof(unsigned long), 1, fp );
    fread( &biSizeImage, sizeof(unsigned long), 1, fp );
    fread( &biXPelsPerMeter, sizeof(long), 1, fp );
    fread( &biYPelsPerMeter, sizeof(long), 1, fp );
    fread( &biClrUsed, sizeof(unsigned long), 1, fp );
    fread( &biClrImportant, sizeof(unsigned long), 1, fp );

    int pad = (biWidth*3+3)/4 * 4 - biWidth*3;
    int imgsize = (biWidth*3 + pad) * biHeight;

    height_map = new unsigned char[biWidth*biHeight];

    unsigned char *buf = new unsigned char[imgsize];
    fread( buf, 1, imgsize, fp );

    int cnt = 0;
    for (int y=biHeight-1; y>=0; y--) {
        for (int x=0; x<biWidth; x++) {
            int data = 0;
            int idx = y*(biWidth*3+pad) + x*3;
            data += buf[idx+0];
            data += buf[idx+1];
            data += buf[idx+2];
            data /= 3; // average
            height_map[cnt] = data;
            cnt++;
        }
    }
    delete [] buf;

    fclose( fp );

    *x_size = biWidth;
    *y_size = biHeight;

    return height_map;
}


dReal heightfield_callback( void* data_p, int x, int y )
{
    unsigned char *map = (unsigned char*)data_p;
    return (dReal)(map[x+y*height_xsize])/64.0;
}


// Փˌop֐
static void nearCallback( void *data, dGeomID o1, dGeomID o2 )
{
	const int MAX_CONTACTS = 10;

	dBodyID b1 = dGeomGetBody( o1 ); // 1
	dBodyID b2 = dGeomGetBody( o2 ); // 2

	if ( b1 && b2 && dAreConnectedExcluding( b1, b2, dJointTypeContact ) )
	return; // ՓˑΏۂłȂ̂̏Փ˂͂͂

	dContact contact[MAX_CONTACTS];
	for ( int i=0; i<MAX_CONTACTS; i++ )
	{
        // ̓m̐ڐG̃p[^ݒ
        contact[i].surface.mode = dContactBounce | dContactSoftCFM | dContactSoftERP;
        contact[i].surface.mu = dInfinity;   // CW
        contact[i].surface.mu2 = 0;
        contact[i].surface.bounce = 0.5;     // W
        contact[i].surface.bounce_vel = 0.1;
        contact[i].surface.soft_cfm = 0.0001;  // CFMݒ
        contact[i].surface.soft_erp = 0.8;
	}

	// Փˌo
	int numc = dCollide( o1, o2, MAX_CONTACTS, &contact[0].geom, sizeof( dContact ) );
	if ( numc > 0 )
	{
		for ( int i=0; i<numc; i++ )
		{
			// Փ˂̔
			dJointID c = dJointCreateContact( world, contactgroup, contact+i );
			dJointAttach( c, b1, b2 );
		}
	}
}



// start simulation - set viewpoint
static void start()
{
  static float xyz[3] = { 0.f, -12.f, 10.f };
  static float hpr[3] = { 90.f, -30.f, 0.f };

  dsSetViewpoint( xyz, hpr );
}

// simulation loop
static void simLoop( int pause )
{
    if (!pause)
    {
        dSpaceCollide( space, 0, &nearCallback );

        dWorldStep( world, 0.01 );

        dJointGroupEmpty( contactgroup );
    }
    // draw ball
    {
	    const dReal* pos = dGeomGetPosition( geom1 );
	    const dReal* R = dGeomGetRotation( geom1 );
        dsSetColor( 1.0, 0.0, 0.0 );
        dsDrawSphere( pos, R, 0.5 );
    }
    
    // draw Heightfield
	{
        const dReal wsamp = ( HFIELD_WIDTH / ( height_xsize-1 ) );
        const dReal dsamp = ( HFIELD_DEPTH / ( height_ysize-1 ) );
	    const dReal* pReal = dGeomGetPosition( gheight );
	    const dReal* RReal = dGeomGetRotation( gheight );
	    // Set ox and oz to zero for DHEIGHTFIELD_CORNER_ORIGIN mode.
	    int ox = (int) ( -(int)HFIELD_WIDTH/2 );
	    int oz = (int) ( -(int)HFIELD_DEPTH/2 );

		dsSetColorAlpha (0.5,1,0.5,0.5);
		dsSetTexture( DS_WOOD );

		for ( int i = 0; i < height_xsize - 1; ++i )
		for ( int j = 0; j < height_ysize - 1; ++j )
		{
			dReal a[3], b[3], c[3], d[3];

			a[ 0 ] = ox + ( i ) * wsamp;
			a[ 1 ] = heightfield_callback( height_map, i, j );
			a[ 2 ] = oz + ( j ) * dsamp;

			b[ 0 ] = ox + ( i + 1 ) * wsamp;
			b[ 1 ] = heightfield_callback( height_map, i + 1, j );
			b[ 2 ] = oz + ( j ) * dsamp;

			c[ 0 ] = ox + ( i ) * wsamp;
			c[ 1 ] = heightfield_callback( height_map, i, j + 1 );
			c[ 2 ] = oz + ( j + 1 ) * dsamp;

			d[ 0 ] = ox + ( i + 1 ) * wsamp;
			d[ 1 ] = heightfield_callback( height_map, i + 1, j + 1 );
			d[ 2 ] = oz + ( j + 1 ) * dsamp;

			dsDrawTriangle( pReal, RReal, a, c, b, 1 );
			dsDrawTriangle( pReal, RReal, b, c, d, 1 );
		}
	}
}


int main( int argc, char* argv[] )
{
    // setup pointers to drawstuff callback functions
    dsFunctions fn;
    fn.version = DS_VERSION;
    fn.start   = &start;
    fn.step    = &simLoop;
    fn.command = 0;
    fn.stop    = 0;
    fn.path_to_textures = "../drawstuff/textures";

    dInitODE();

    // create world
    world = dWorldCreate();
    space = dHashSpaceCreate( 0 );
    contactgroup = dJointGroupCreate( 0 );
    dCreatePlane( space, 0, 0, 1, 0 );
    dWorldSetCFM (world,1e-5);
    dWorldSetERP (world,0.1);
    dWorldSetGravity( world, 0.0, 0.0, -9.8 );

    height_map = const_cast<unsigned char*>( createHeightMapForBitmap( "a.bmp", &height_xsize, &height_ysize ) );
    if (!height_map) {
        printf("NG\n");
        exit(1);
    }

    dHeightfieldDataID heightid = dGeomHeightfieldDataCreate();
    dGeomHeightfieldDataBuildCallback(
        heightid,
        (void*)height_map,
        heightfield_callback,
        HFIELD_WIDTH,  // width
        HFIELD_DEPTH,  // depth
        height_xsize, // width steps
        height_ysize, // depth steps
        1.0,
        0.0,
        0.0,
        0
    );

	dGeomHeightfieldDataSetBounds( heightid, REAL( -4.0 ), REAL( +6.0 ) );

	gheight = dCreateHeightfield( space, heightid, 1 );

	dVector3 pos;
	pos[ 0 ] = 0;
	pos[ 1 ] = 0;
	pos[ 2 ] = 0.1;

	// Rotate so Z is up, not Y (which is the default orientation)
	dMatrix3 R;
	dRSetIdentity( R );
	dRFromAxisAndAngle( R, 1, 0, 0, 90.0*M_PI/180.0 );

	// Place it.
	dGeomSetRotation( gheight, R );
	dGeomSetPosition( gheight, pos[0], pos[1], pos[2] );

    {
        // ball
        dReal r = 0.5;
	    body1 = dBodyCreate( world );
	    geom1 = dCreateSphere( space, r );

        dMass m;
	    dMassSetSphere( &m, 1.0f, r );
	    dMassAdjust( &m, 1.0 );
	    dGeomSetBody( geom1, body1 );

        dBodySetPosition( body1, 1.0, -1.5, 10.0 );
    }


    // starting simulation
    dsSimulationLoop( argc, argv, 320, 240, &fn );

    dJointGroupDestroy( contactgroup );
    dSpaceDestroy( space );
    dWorldDestroy( world );
    return 0;
}

